-module(spoke@internal@packet@client@incoming).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([decode_packet/1]).
-export_type([packet/0]).

-if(?OTP_RELEASE >= 27).
-define(MODULEDOC(Str), -moduledoc(Str)).
-define(DOC(Str), -doc(Str)).
-else.
-define(MODULEDOC(Str), -compile([])).
-define(DOC(Str), -compile([])).
-endif.

?MODULEDOC(false).

-type packet() :: {conn_ack,
        {ok, boolean()} | {error, spoke@internal@packet:connect_error()}} |
    {publish, spoke@internal@packet:publish_data()} |
    {pub_ack, integer()} |
    {pub_rec, integer()} |
    {pub_rel, integer()} |
    {pub_comp, integer()} |
    {sub_ack,
        integer(),
        list({ok, spoke@internal@packet:qo_s()} | {error, nil})} |
    {unsub_ack, integer()} |
    ping_resp.

-file("src/spoke/internal/packet/client/incoming.gleam", 20).
?DOC(false).
-spec decode_packet(bitstring()) -> {ok, {packet(), bitstring()}} |
    {error, spoke@internal@packet@decode:decode_error()}.
decode_packet(Bytes) ->
    case Bytes of
        <<Id:4, Flags:4/bitstring, Rest/binary>> ->
            case Id of
                2 ->
                    spoke@internal@packet@decode:connack(
                        Flags,
                        Rest,
                        fun(Field@0) -> {conn_ack, Field@0} end
                    );

                3 ->
                    spoke@internal@packet@decode:publish(
                        Flags,
                        Rest,
                        fun(Field@0) -> {publish, Field@0} end
                    );

                4 ->
                    spoke@internal@packet@decode:only_packet_id(
                        Flags,
                        Rest,
                        0,
                        fun(Field@0) -> {pub_ack, Field@0} end
                    );

                5 ->
                    spoke@internal@packet@decode:only_packet_id(
                        Flags,
                        Rest,
                        0,
                        fun(Field@0) -> {pub_rec, Field@0} end
                    );

                6 ->
                    spoke@internal@packet@decode:only_packet_id(
                        Flags,
                        Rest,
                        2,
                        fun(Field@0) -> {pub_rel, Field@0} end
                    );

                7 ->
                    spoke@internal@packet@decode:only_packet_id(
                        Flags,
                        Rest,
                        0,
                        fun(Field@0) -> {pub_comp, Field@0} end
                    );

                9 ->
                    spoke@internal@packet@decode:suback(
                        Flags,
                        Rest,
                        fun(Field@0, Field@1) -> {sub_ack, Field@0, Field@1} end
                    );

                11 ->
                    spoke@internal@packet@decode:only_packet_id(
                        Flags,
                        Rest,
                        0,
                        fun(Field@0) -> {unsub_ack, Field@0} end
                    );

                13 ->
                    spoke@internal@packet@decode:zero_length(
                        Flags,
                        Rest,
                        ping_resp
                    );

                _ ->
                    {error, {invalid_packet_identifier, Id}}
            end;

        _ ->
            {error, data_too_short}
    end.
