-module(spoke@internal@packet@client@outgoing).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([encode_packet/1]).
-export_type([packet/0]).

-if(?OTP_RELEASE >= 27).
-define(MODULEDOC(Str), -moduledoc(Str)).
-define(DOC(Str), -doc(Str)).
-else.
-define(MODULEDOC(Str), -compile([])).
-define(DOC(Str), -compile([])).
-endif.

?MODULEDOC(false).

-type packet() :: {connect, spoke@internal@packet:connect_options()} |
    {publish, spoke@internal@packet:publish_data()} |
    {pub_ack, integer()} |
    {pub_rec, integer()} |
    {pub_rel, integer()} |
    {pub_comp, integer()} |
    {subscribe, integer(), list(spoke@internal@packet:subscribe_request())} |
    {unsubscribe, integer(), list(binary())} |
    ping_req |
    disconnect.

-file("src/spoke/internal/packet/client/outgoing.gleam", 22).
?DOC(false).
-spec encode_packet(packet()) -> {ok, gleam@bytes_tree:bytes_tree()} |
    {error, spoke@internal@packet@encode:encode_error()}.
encode_packet(Packet) ->
    case Packet of
        {connect, Data} ->
            {ok, spoke@internal@packet@encode:connect(Data)};

        {publish, Data@1} ->
            {ok, spoke@internal@packet@encode:publish(Data@1)};

        {pub_ack, Packet_id} ->
            {ok, spoke@internal@packet@encode:only_packet_id(4, 0, Packet_id)};

        {pub_rec, Packet_id@1} ->
            {ok, spoke@internal@packet@encode:only_packet_id(5, 0, Packet_id@1)};

        {pub_rel, Packet_id@2} ->
            {ok, spoke@internal@packet@encode:only_packet_id(6, 2, Packet_id@2)};

        {pub_comp, Packet_id@3} ->
            {ok, spoke@internal@packet@encode:only_packet_id(7, 0, Packet_id@3)};

        {subscribe, Packet_id@4, Topics} ->
            spoke@internal@packet@encode:subscribe(Packet_id@4, Topics);

        {unsubscribe, Packet_id@5, Topics@1} ->
            spoke@internal@packet@encode:unsubscribe(Packet_id@5, Topics@1);

        ping_req ->
            {ok, spoke@internal@packet@encode:only_packet_type(12)};

        disconnect ->
            {ok, spoke@internal@packet@encode:only_packet_type(14)}
    end.
