-module(spoke@internal@packet@server@incoming).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([decode_packet/1]).
-export_type([packet/0]).

-if(?OTP_RELEASE >= 27).
-define(MODULEDOC(Str), -moduledoc(Str)).
-define(DOC(Str), -doc(Str)).
-else.
-define(MODULEDOC(Str), -compile([])).
-define(DOC(Str), -compile([])).
-endif.

?MODULEDOC(false).

-type packet() :: {connect, spoke@internal@packet:connect_options()} |
    {publish, spoke@internal@packet:publish_data()} |
    {pub_ack, integer()} |
    {pub_rec, integer()} |
    {pub_rel, integer()} |
    {pub_comp, integer()} |
    {subscribe, integer(), list(spoke@internal@packet:subscribe_request())} |
    {unsubscribe, integer(), list(binary())} |
    ping_req |
    disconnect.

-file("src/spoke/internal/packet/server/incoming.gleam", 23).
?DOC(false).
-spec decode_packet(bitstring()) -> {ok, {packet(), bitstring()}} |
    {error, spoke@internal@packet@decode:decode_error()}.
decode_packet(Bytes) ->
    case Bytes of
        <<Id:4, Flags:4/bitstring, Rest/binary>> ->
            case Id of
                1 ->
                    spoke@internal@packet@decode:connect(
                        Flags,
                        Rest,
                        fun(Field@0) -> {connect, Field@0} end
                    );

                3 ->
                    spoke@internal@packet@decode:publish(
                        Flags,
                        Rest,
                        fun(Field@0) -> {publish, Field@0} end
                    );

                4 ->
                    spoke@internal@packet@decode:only_packet_id(
                        Flags,
                        Rest,
                        0,
                        fun(Field@0) -> {pub_ack, Field@0} end
                    );

                5 ->
                    spoke@internal@packet@decode:only_packet_id(
                        Flags,
                        Rest,
                        0,
                        fun(Field@0) -> {pub_rec, Field@0} end
                    );

                6 ->
                    spoke@internal@packet@decode:only_packet_id(
                        Flags,
                        Rest,
                        2,
                        fun(Field@0) -> {pub_rel, Field@0} end
                    );

                7 ->
                    spoke@internal@packet@decode:only_packet_id(
                        Flags,
                        Rest,
                        0,
                        fun(Field@0) -> {pub_comp, Field@0} end
                    );

                8 ->
                    spoke@internal@packet@decode:subscribe(
                        Flags,
                        Rest,
                        fun(Field@0, Field@1) -> {subscribe, Field@0, Field@1} end
                    );

                10 ->
                    spoke@internal@packet@decode:unsubscribe(
                        Flags,
                        Rest,
                        fun(Field@0, Field@1) -> {unsubscribe, Field@0, Field@1} end
                    );

                12 ->
                    spoke@internal@packet@decode:zero_length(
                        Flags,
                        Rest,
                        ping_req
                    );

                14 ->
                    spoke@internal@packet@decode:zero_length(
                        Flags,
                        Rest,
                        disconnect
                    );

                _ ->
                    {error, {invalid_packet_identifier, Id}}
            end;

        _ ->
            {error, data_too_short}
    end.
