-module(spoke@internal@packet@server@outgoing).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([encode_packet/1]).
-export_type([packet/0]).

-if(?OTP_RELEASE >= 27).
-define(MODULEDOC(Str), -moduledoc(Str)).
-define(DOC(Str), -doc(Str)).
-else.
-define(MODULEDOC(Str), -compile([])).
-define(DOC(Str), -compile([])).
-endif.

?MODULEDOC(false).

-type packet() :: {conn_ack,
        {ok, boolean()} | {error, spoke@internal@packet:connect_error()}} |
    {publish, spoke@internal@packet:publish_data()} |
    {pub_ack, integer()} |
    {pub_rec, integer()} |
    {pub_rel, integer()} |
    {pub_comp, integer()} |
    {sub_ack,
        integer(),
        list({ok, spoke@internal@packet:qo_s()} | {error, nil})} |
    {unsub_ack, integer()} |
    ping_resp.

-file("src/spoke/internal/packet/server/outgoing.gleam", 23).
?DOC(false).
-spec encode_packet(packet()) -> {ok, gleam@bytes_tree:bytes_tree()} |
    {error, spoke@internal@packet@encode:encode_error()}.
encode_packet(Packet) ->
    case Packet of
        {conn_ack, Result} ->
            {ok, spoke@internal@packet@encode:connack(Result)};

        {publish, Data} ->
            {ok, spoke@internal@packet@encode:publish(Data)};

        {pub_ack, Packet_id} ->
            {ok, spoke@internal@packet@encode:only_packet_id(4, 0, Packet_id)};

        {pub_rec, Packet_id@1} ->
            {ok, spoke@internal@packet@encode:only_packet_id(5, 0, Packet_id@1)};

        {pub_rel, Packet_id@2} ->
            {ok, spoke@internal@packet@encode:only_packet_id(6, 2, Packet_id@2)};

        {pub_comp, Packet_id@3} ->
            {ok, spoke@internal@packet@encode:only_packet_id(7, 0, Packet_id@3)};

        {sub_ack, Packet_id@4, Results} ->
            spoke@internal@packet@encode:suback(Packet_id@4, Results);

        {unsub_ack, Packet_id@5} ->
            {ok,
                spoke@internal@packet@encode:only_packet_id(11, 0, Packet_id@5)};

        ping_resp ->
            {ok, spoke@internal@packet@encode:only_packet_type(13)}
    end.
