-module(spoke@internal@session).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([new/1, from_json/1, is_ephemeral/1, pending_publishes/1, reserve_packet_id/1, start_qos1_publish/2, start_qos2_publish/2, handle_pubrec/2, handle_pubcomp/2, start_qos2_receive/2, handle_puback/2, handle_pubrel/2, packets_to_send_after_connect/1, to_json/1]).
-export_type([session/0, pub_ack_result/0]).

-if(?OTP_RELEASE >= 27).
-define(MODULEDOC(Str), -moduledoc(Str)).
-define(DOC(Str), -doc(Str)).
-else.
-define(MODULEDOC(Str), -compile([])).
-define(DOC(Str), -compile([])).
-endif.

?MODULEDOC(false).

-opaque session() :: {session,
        boolean(),
        integer(),
        gleam@dict:dict(integer(), spoke@internal@packet:message_data()),
        gleam@dict:dict(integer(), spoke@internal@packet:message_data()),
        gleam@set:set(integer()),
        gleam@set:set(integer())}.

-type pub_ack_result() :: {publish_finished, session()} | invalid_pub_ack_id.

-file("src/spoke/internal/session.gleam", 47).
?DOC(false).
-spec new(boolean()) -> session().
new(Clean_session) ->
    {session,
        Clean_session,
        1,
        maps:new(),
        maps:new(),
        gleam@set:new(),
        gleam@set:new()}.

-file("src/spoke/internal/session.gleam", 58).
?DOC(false).
-spec from_json(binary()) -> {ok, session()} |
    {error, gleam@json:decode_error()}.
from_json(State) ->
    Decode_payload = begin
        _pipe = {decoder, fun gleam@dynamic@decode:decode_string/1},
        gleam@dynamic@decode:then(
            _pipe,
            fun(Str) -> case gleam@bit_array:base64_decode(Str) of
                    {ok, Payload} ->
                        gleam@dynamic@decode:success(Payload);

                    {error, nil} ->
                        gleam@dynamic@decode:failure(<<>>, <<"BitArray"/utf8>>)
                end end
        )
    end,
    Decode_id_key = begin
        _pipe@1 = {decoder, fun gleam@dynamic@decode:decode_string/1},
        gleam@dynamic@decode:then(
            _pipe@1,
            fun(Str@1) -> case gleam_stdlib:parse_int(Str@1) of
                    {ok, Id} ->
                        gleam@dynamic@decode:success(Id);

                    {error, nil} ->
                        gleam@dynamic@decode:failure(0, <<"Int"/utf8>>)
                end end
        )
    end,
    Decode_message = begin
        gleam@dynamic@decode:field(
            <<"topic"/utf8>>,
            {decoder, fun gleam@dynamic@decode:decode_string/1},
            fun(Topic) ->
                gleam@dynamic@decode:field(
                    <<"payload"/utf8>>,
                    Decode_payload,
                    fun(Payload@1) ->
                        gleam@dynamic@decode:field(
                            <<"retain"/utf8>>,
                            {decoder, fun gleam@dynamic@decode:decode_bool/1},
                            fun(Retain) ->
                                gleam@dynamic@decode:success(
                                    {message_data, Topic, Payload@1, Retain}
                                )
                            end
                        )
                    end
                )
            end
        )
    end,
    Decode_message_dict = gleam@dynamic@decode:dict(
        Decode_id_key,
        Decode_message
    ),
    Decode_id_set = begin
        _pipe@2 = gleam@dynamic@decode:list(
            {decoder, fun gleam@dynamic@decode:decode_int/1}
        ),
        gleam@dynamic@decode:map(_pipe@2, fun gleam@set:from_list/1)
    end,
    Core_decoder = begin
        gleam@dynamic@decode:field(
            <<"packet_id"/utf8>>,
            {decoder, fun gleam@dynamic@decode:decode_int/1},
            fun(Packet_id) ->
                gleam@dynamic@decode:field(
                    <<"unacked_qos1"/utf8>>,
                    Decode_message_dict,
                    fun(Unacked_qos1) ->
                        gleam@dynamic@decode:field(
                            <<"unreceived_qos2"/utf8>>,
                            Decode_message_dict,
                            fun(Unreceived_qos2) ->
                                gleam@dynamic@decode:field(
                                    <<"unreleased_qos2"/utf8>>,
                                    Decode_id_set,
                                    fun(Unreleased_qos2) ->
                                        gleam@dynamic@decode:field(
                                            <<"incomplete_qos2_in"/utf8>>,
                                            Decode_id_set,
                                            fun(Incomplete_qos2_in) ->
                                                gleam@dynamic@decode:success(
                                                    {session,
                                                        false,
                                                        Packet_id,
                                                        Unacked_qos1,
                                                        Unreceived_qos2,
                                                        Unreleased_qos2,
                                                        Incomplete_qos2_in}
                                                )
                                            end
                                        )
                                    end
                                )
                            end
                        )
                    end
                )
            end
        )
    end,
    Decoder = begin
        _pipe@3 = begin
            gleam@dynamic@decode:field(
                <<"version"/utf8>>,
                {decoder, fun gleam@dynamic@decode:decode_int/1},
                fun(Version) -> case Version of
                        1 ->
                            gleam@dynamic@decode:success(Version);

                        _ ->
                            gleam@dynamic@decode:failure(
                                Version,
                                <<"known version"/utf8>>
                            )
                    end end
            )
        end,
        gleam@dynamic@decode:then(
            _pipe@3,
            fun(_) ->
                begin
                    gleam@dynamic@decode:field(
                        <<"data"/utf8>>,
                        gleam@dynamic@decode:optional(Core_decoder),
                        fun(Session) ->
                            gleam@dynamic@decode:success(
                                gleam@option:unwrap(Session, new(true))
                            )
                        end
                    )
                end
            end
        )
    end,
    gleam@json:parse(State, Decoder).

-file("src/spoke/internal/session.gleam", 161).
?DOC(false).
-spec is_ephemeral(session()) -> boolean().
is_ephemeral(Session) ->
    erlang:element(2, Session).

-file("src/spoke/internal/session.gleam", 165).
?DOC(false).
-spec pending_publishes(session()) -> integer().
pending_publishes(Session) ->
    (maps:size(erlang:element(4, Session)) + maps:size(
        erlang:element(5, Session)
    ))
    + gleam@set:size(erlang:element(6, Session)).

-file("src/spoke/internal/session.gleam", 171).
?DOC(false).
-spec reserve_packet_id(session()) -> {session(), integer()}.
reserve_packet_id(Session) ->
    Id = erlang:element(3, Session),
    Next_id = case Id of
        65535 ->
            1;

        Id@1 ->
            Id@1 + 1
    end,
    {begin
            _record = Session,
            {session,
                erlang:element(2, _record),
                Next_id,
                erlang:element(4, _record),
                erlang:element(5, _record),
                erlang:element(6, _record),
                erlang:element(7, _record)}
        end,
        Id}.

-file("src/spoke/internal/session.gleam", 182).
?DOC(false).
-spec start_qos1_publish(session(), spoke@internal@packet:message_data()) -> {session(),
    spoke@internal@packet@client@outgoing:packet()}.
start_qos1_publish(Session, Message) ->
    {Session@1, Id} = reserve_packet_id(Session),
    Unacked_qos1 = gleam@dict:insert(erlang:element(4, Session@1), Id, Message),
    Packet = {publish, {publish_data_qo_s1, Message, false, Id}},
    {begin
            _record = Session@1,
            {session,
                erlang:element(2, _record),
                erlang:element(3, _record),
                Unacked_qos1,
                erlang:element(5, _record),
                erlang:element(6, _record),
                erlang:element(7, _record)}
        end,
        Packet}.

-file("src/spoke/internal/session.gleam", 195).
?DOC(false).
-spec start_qos2_publish(session(), spoke@internal@packet:message_data()) -> {session(),
    spoke@internal@packet@client@outgoing:packet()}.
start_qos2_publish(Session, Message) ->
    {Session@1, Id} = reserve_packet_id(Session),
    Unreceived_qos2 = gleam@dict:insert(
        erlang:element(5, Session@1),
        Id,
        Message
    ),
    Packet = {publish, {publish_data_qo_s2, Message, false, Id}},
    {begin
            _record = Session@1,
            {session,
                erlang:element(2, _record),
                erlang:element(3, _record),
                erlang:element(4, _record),
                Unreceived_qos2,
                erlang:element(6, _record),
                erlang:element(7, _record)}
        end,
        Packet}.

-file("src/spoke/internal/session.gleam", 208).
?DOC(false).
-spec handle_pubrec(session(), integer()) -> session().
handle_pubrec(Session, Id) ->
    Unreceived_qos2 = gleam@dict:delete(erlang:element(5, Session), Id),
    Unreleased_qos2 = gleam@set:insert(erlang:element(6, Session), Id),
    _record = Session,
    {session,
        erlang:element(2, _record),
        erlang:element(3, _record),
        erlang:element(4, _record),
        Unreceived_qos2,
        Unreleased_qos2,
        erlang:element(7, _record)}.

-file("src/spoke/internal/session.gleam", 214).
?DOC(false).
-spec handle_pubcomp(session(), integer()) -> session().
handle_pubcomp(Session, Id) ->
    Unreleased_qos2 = gleam@set:delete(erlang:element(6, Session), Id),
    _record = Session,
    {session,
        erlang:element(2, _record),
        erlang:element(3, _record),
        erlang:element(4, _record),
        erlang:element(5, _record),
        Unreleased_qos2,
        erlang:element(7, _record)}.

-file("src/spoke/internal/session.gleam", 222).
?DOC(false).
-spec start_qos2_receive(session(), integer()) -> {session(), boolean()}.
start_qos2_receive(Session, Packet_id) ->
    case gleam@set:contains(erlang:element(7, Session), Packet_id) of
        true ->
            {Session, false};

        false ->
            Incomplete_qos2_in = gleam@set:insert(
                erlang:element(7, Session),
                Packet_id
            ),
            {begin
                    _record = Session,
                    {session,
                        erlang:element(2, _record),
                        erlang:element(3, _record),
                        erlang:element(4, _record),
                        erlang:element(5, _record),
                        erlang:element(6, _record),
                        Incomplete_qos2_in}
                end,
                true}
    end.

-file("src/spoke/internal/session.gleam", 232).
?DOC(false).
-spec handle_puback(session(), integer()) -> pub_ack_result().
handle_puback(Session, Packet_id) ->
    case gleam@dict:has_key(erlang:element(4, Session), Packet_id) of
        true ->
            Unacked_qos1 = gleam@dict:delete(
                erlang:element(4, Session),
                Packet_id
            ),
            {publish_finished,
                begin
                    _record = Session,
                    {session,
                        erlang:element(2, _record),
                        erlang:element(3, _record),
                        Unacked_qos1,
                        erlang:element(5, _record),
                        erlang:element(6, _record),
                        erlang:element(7, _record)}
                end};

        false ->
            invalid_pub_ack_id
    end.

-file("src/spoke/internal/session.gleam", 242).
?DOC(false).
-spec handle_pubrel(session(), integer()) -> session().
handle_pubrel(Session, Packet_id) ->
    Incomplete_qos2_in = gleam@set:delete(erlang:element(7, Session), Packet_id),
    _record = Session,
    {session,
        erlang:element(2, _record),
        erlang:element(3, _record),
        erlang:element(4, _record),
        erlang:element(5, _record),
        erlang:element(6, _record),
        Incomplete_qos2_in}.

-file("src/spoke/internal/session.gleam", 247).
?DOC(false).
-spec packets_to_send_after_connect(session()) -> list(spoke@internal@packet@client@outgoing:packet()).
packets_to_send_after_connect(Session) ->
    Packets = gleam@list:new(),
    Packets@2 = gleam@list:fold(
        maps:keys(erlang:element(4, Session)),
        Packets,
        fun(Packets@1, Id) ->
            _assert_subject = gleam_stdlib:map_get(
                erlang:element(4, Session),
                Id
            ),
            {ok, Message} = case _assert_subject of
                {ok, _} -> _assert_subject;
                _assert_fail ->
                    erlang:error(#{gleam_error => let_assert,
                                message => <<"Pattern match failed, no pattern matched the value."/utf8>>,
                                value => _assert_fail,
                                module => <<"spoke/internal/session"/utf8>>,
                                function => <<"packets_to_send_after_connect"/utf8>>,
                                line => 253})
            end,
            Packet = {publish, {publish_data_qo_s1, Message, true, Id}},
            [Packet | Packets@1]
        end
    ),
    Packets@4 = gleam@list:fold(
        maps:keys(erlang:element(5, Session)),
        Packets@2,
        fun(Packets@3, Id@1) ->
            _assert_subject@1 = gleam_stdlib:map_get(
                erlang:element(5, Session),
                Id@1
            ),
            {ok, Message@1} = case _assert_subject@1 of
                {ok, _} -> _assert_subject@1;
                _assert_fail@1 ->
                    erlang:error(#{gleam_error => let_assert,
                                message => <<"Pattern match failed, no pattern matched the value."/utf8>>,
                                value => _assert_fail@1,
                                module => <<"spoke/internal/session"/utf8>>,
                                function => <<"packets_to_send_after_connect"/utf8>>,
                                line => 266})
            end,
            Packet@1 = {publish, {publish_data_qo_s2, Message@1, true, Id@1}},
            [Packet@1 | Packets@3]
        end
    ),
    Packets@6 = gleam@list:fold(
        gleam@set:to_list(erlang:element(6, Session)),
        Packets@4,
        fun(Packets@5, Id@2) -> [{pub_rel, Id@2} | Packets@5] end
    ),
    Packets@6.

-file("src/spoke/internal/session.gleam", 123).
?DOC(false).
-spec to_json(session()) -> binary().
to_json(Session) ->
    Encode_message = fun(Msg) ->
        gleam@json:object(
            [{<<"topic"/utf8>>, gleam@json:string(erlang:element(2, Msg))},
                {<<"payload"/utf8>>,
                    gleam@json:string(
                        gleam_stdlib:bit_array_base64_encode(
                            erlang:element(3, Msg),
                            true
                        )
                    )},
                {<<"retain"/utf8>>, gleam@json:bool(erlang:element(4, Msg))}]
        )
    end,
    Encode_message_dict = fun(_capture) ->
        gleam@json:dict(
            _capture,
            fun erlang:integer_to_binary/1,
            Encode_message
        )
    end,
    Encode_id_set = fun(Ids) ->
        gleam@json:array(gleam@set:to_list(Ids), fun gleam@json:int/1)
    end,
    Data = case Session of
        {session,
            false,
            Packet_id,
            Unacked_qos1,
            Unreceived_qos2,
            Unreleased_qos2,
            Incomplete_qos2_in} ->
            {some,
                [{<<"packet_id"/utf8>>, gleam@json:int(Packet_id)},
                    {<<"unacked_qos1"/utf8>>, Encode_message_dict(Unacked_qos1)},
                    {<<"unreceived_qos2"/utf8>>,
                        Encode_message_dict(Unreceived_qos2)},
                    {<<"unreleased_qos2"/utf8>>, Encode_id_set(Unreleased_qos2)},
                    {<<"incomplete_qos2_in"/utf8>>,
                        Encode_id_set(Incomplete_qos2_in)}]};

        _ ->
            none
    end,
    _pipe = gleam@json:object(
        [{<<"version"/utf8>>, gleam@json:int(1)},
            {<<"data"/utf8>>,
                gleam@json:nullable(Data, fun gleam@json:object/1)}]
    ),
    gleam@json:to_string(_pipe).
