-module(spoke@internal@transport).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([encode_channel/1]).
-export_type([byte_channel/0, encoded_channel/0]).

-if(?OTP_RELEASE >= 27).
-define(MODULEDOC(Str), -moduledoc(Str)).
-define(DOC(Str), -doc(Str)).
-else.
-define(MODULEDOC(Str), -compile([])).
-define(DOC(Str), -compile([])).
-endif.

?MODULEDOC(false).

-type byte_channel() :: {byte_channel,
        fun((gleam@bytes_tree:bytes_tree()) -> {ok, nil} | {error, binary()}),
        fun(() -> gleam@erlang@process:selector({ok, bitstring()} |
            {error, binary()})),
        fun(() -> nil)}.

-type encoded_channel() :: {encoded_channel,
        fun((spoke@internal@packet@client@outgoing:packet()) -> {ok, nil} |
            {error, binary()}),
        fun((bitstring()) -> gleam@erlang@process:selector({bitstring(),
            {ok, list(spoke@internal@packet@client@incoming:packet())} |
                {error, binary()}})),
        fun(() -> nil)}.

-file("src/spoke/internal/transport.gleam", 47).
?DOC(false).
-spec send(byte_channel(), spoke@internal@packet@client@outgoing:packet()) -> {ok,
        nil} |
    {error, binary()}.
send(Channel, Packet) ->
    case spoke@internal@packet@client@outgoing:encode_packet(Packet) of
        {ok, Bytes} ->
            (erlang:element(2, Channel))(Bytes);

        {error, E} ->
            {error, <<"Encode error: "/utf8, (gleam@string:inspect(E))/binary>>}
    end.

-file("src/spoke/internal/transport.gleam", 69).
?DOC(false).
-spec decode(bitstring(), list(spoke@internal@packet@client@incoming:packet())) -> {ok,
        {bitstring(), list(spoke@internal@packet@client@incoming:packet())}} |
    {error, binary()}.
decode(Data, Packets) ->
    case spoke@internal@packet@client@incoming:decode_packet(Data) of
        {error, data_too_short} ->
            {ok, {Data, Packets}};

        {error, E} ->
            {error, <<"Invalid data: "/utf8, (gleam@string:inspect(E))/binary>>};

        {ok, {Packet, Rest}} ->
            decode(Rest, [Packet | Packets])
    end.

-file("src/spoke/internal/transport.gleam", 54).
?DOC(false).
-spec selecting_next(bitstring(), byte_channel()) -> gleam@erlang@process:selector({bitstring(),
    {ok, list(spoke@internal@packet@client@incoming:packet())} |
        {error, binary()}}).
selecting_next(State, Channel) ->
    gleam_erlang_ffi:map_selector(
        (erlang:element(3, Channel))(),
        fun(Input) -> case Input of
                {ok, Data} ->
                    case decode(gleam@bit_array:append(State, Data), []) of
                        {ok, {Rest, Result}} ->
                            {Rest, {ok, lists:reverse(Result)}};

                        {error, E} ->
                            {<<>>,
                                {error,
                                    <<"Invalid data: "/utf8,
                                        (gleam@string:inspect(E))/binary>>}}
                    end;

                {error, E@1} ->
                    {State, {error, E@1}}
            end end
    ).

-file("src/spoke/internal/transport.gleam", 39).
?DOC(false).
-spec encode_channel(byte_channel()) -> encoded_channel().
encode_channel(Channel) ->
    {encoded_channel,
        fun(_capture) -> send(Channel, _capture) end,
        fun(_capture@1) -> selecting_next(_capture@1, Channel) end,
        erlang:element(4, Channel)}.
